function [ results , clusters ] = knn_overlappinggrid( database_orig , k, datarep, defaultNonDetectedValue, newNonDetectedValue , distanceMetric , k2, randseed)
% knn_ips Indoor Positioning System based on kNN
%   
% inputs: 
%    database_orig           : database with RSSI values and Labels
%    k                       : value of nearest neighbors in kNN algoritm
%    datarep                 : Data representation
%    defaultNonDetectedValue : Value in DB for non detected RSSI values
%    newNonDetectedValue     : New Value for non detected RSSI values
%    distanceMetric          : Distance metric used for FP matching
%
% output:
%    results                 : output statistics
%
% Developed by J. Torres-Sospedra,
% Instiute of New Imaging Technologies, Universitat Jaume I
% jtorres@uji.es

% Load for usage with Octave
IS_OCTAVE = exist('OCTAVE_VERSION', 'builtin') ~= 0;
if IS_OCTAVE
    pkg load statistics
    more off;
end

ticT1m3 = myTic();
% Remap bld numbers to 1:nblds
origBlds = unique((database_orig.trainingLabels(:,5))); 
nblds    = size(unique(origBlds),1);
database0  = remapBldDB(database_orig,origBlds',1:nblds);

% Remap floors numbers to 1:nfloors
origFloors = unique((database_orig.trainingLabels(:,4))); 
nfloors    = size(unique(origFloors),1);
database0  = remapFloorDB(database0,origFloors',1:nfloors);

% Calculate the overall min RSSI (minus 1)
if size(newNonDetectedValue,1) == 0
newNonDetectedValue = min(...
                     [database0.trainingMacs(:)',...                      
                      database0.testMacs(:)']...
                     )-1;
end

% Replace NonDetectedValue
if size(defaultNonDetectedValue,1) ~= 0
    database0 = datarepNewNullDB(database0,defaultNonDetectedValue,newNonDetectedValue);
end
database_predata = database0;
% Apply new data representation                 
if  strcmp(datarep,'positive') % Convert negative values to positive by adding a value
    database  = datarepPositive(database0);
  
    if or(strcmp(distanceMetric,'distance_PLGD40'),strcmp(distanceMetric,'distance_PLGD40'))
      additionalparams = -85 - newNonDetectedValue;
    else
      additionalparams = 0;
    end
    
elseif strcmp(datarep,'exponential') % Convert to exponential data representation
    %database0 = datarepPositive(database0);
    database  = datarepExponential(database0);
    
    if or(strcmp(distanceMetric,'distance_PLGD40'),strcmp(distanceMetric,'distance_PLGD40'))
      additionalparams =  exp(((-85-newNonDetectedValue))/24)./exp((-newNonDetectedValue)/24);
    else
      additionalparams = 0;
    end
    
elseif strcmp(datarep,'powed') % Convert to powed data representation
    %database0 = datarepPositive(database0); 
    database  = datarepPowed(database0);  
    
    if or(strcmp(distanceMetric,'distance_PLGD40'),strcmp(distanceMetric,'distance_PLGD40'))
      additionalparams =  ((-85-newNonDetectedValue).^exp(1))./((-newNonDetectedValue).^exp(1));
    else
      additionalparams = 0;
    end
      
end % Default, no conversion

clear database0;
   
% Get number of samples
rsamples = size(database.trainingMacs,1); 
osamples = size(database.testMacs,1);
nmacs    = size(database.testMacs,2);

% Prepare output vectors
results.error      = zeros(osamples,5);
results.prediction = zeros(osamples,5);
results.targets    = zeros(osamples,5);
results.candidates = zeros(osamples,1);
results.distances  = zeros(osamples,1);
results.timesample = zeros(osamples,5);
results.considered = zeros(osamples,5);

% Get the deviations values for Mahalanobis Distance
%if strcmp(distanceMetric,'distance_mahalanobis')
    %deviation   = std(database.trainingMacs);
    %deviationsq = (deviation+(10000*deviation==0)).^2;
%end

fprintf('\nRunning the %s (%d) algorithm\n',mfilename)
fprintf('    database features      : [%d,%d,%d]\n',rsamples,osamples,nmacs)
fprintf('    k                      : %d\n',k)
fprintf('    datarep                : %s\n',datarep)
fprintf('    defaultNonDetectedValue: %d\n',defaultNonDetectedValue)
fprintf('    newNonDetectedValue    : %d\n',newNonDetectedValue)
fprintf('    distanceMetric         : %s\n',distanceMetric )

time00pre = myToc(ticT1m3);

database.trainingValidMacs = (database_orig.trainingMacs ~= defaultNonDetectedValue); 
database.testValidMacs     = (database_orig.testMacs     ~= defaultNonDetectedValue);
vecidxmacs    = 1:nmacs;
vecidxsamples = 1:rsamples;

%% Initialize clusters
%clusters.randseed = rsamples+osamples+randseed+103;
%rand('seed',clusters.randseed);

% Get the value of the newNonDetectedValue in the current dataRepresentation
if size(database.trainingMacs(database.trainingValidMacs==0),1) > 0
  newNonDetectedValueProcessed =  mean(database.trainingMacs(database.trainingValidMacs==0));
else
  % if there is not any case where an AP was not heard, we can safely use 0
  newNonDetectedValueProcessed =  0; 
end

% Initialize cluster variables for the whole dataset
clusterGlobal = 0;
clusters.idx       = zeros(rsamples,1);     % Cluster idx for each training sample 
clusters.idx2      = zeros(rsamples,1);     % Cluster idx for each training sample 
clusters.centroids = zeros(rsamples,nmacs); % Grid coordinates for each cluster centroid
clusters.grid      = zeros(rsamples,3);     % Grid coordinates for each cluster centroid

% The properties for the cells are set globally to have the same grids in all floors
minCoord        = min(database.trainingLabels(:,1:2));  % Min and Max coordinates (X-Y only)
maxCoord        = max(database.trainingLabels(:,1:2));
ncells          = ceil([maxCoord-minCoord]/k2)+1;       % Number of Cells per dimension. 
                                                        % Add 1 because of [0...cellsize[ , ... , [(n-1)*cellsize...n*cellsize[
                                                      
for currentFloor = 1:nfloors

  clusterGlobal_previous   = clusterGlobal;
  
  % Reduce radiomap to those fingerprints which belong to the current floor
  idxsfloor                = (database.trainingLabels(:,4) == currentFloor);
  reducedTrainingLabels    = database.trainingLabels(idxsfloor,:);
  reducedTrainingMacs      = database.trainingMacs(idxsfloor,:);
  reducedTrainingValidMacs = database.trainingValidMacs(idxsfloor,:);
  
  rsamplesReduced = size(reducedTrainingLabels,1);      % Number of sampels of the reduced set

  % Initialize cluster variables for current Floor
  cluster = 0;
  clustersFloor{currentFloor}.idx           = zeros(rsamplesReduced,1);           % Cluster idx for each training sample 
  clustersFloor{currentFloor}.idx2          = zeros(rsamplesReduced,1);           % Cluster idx for each training sample   
  clustersFloor{currentFloor}.grid          = zeros(ncells(1)*ncells(2)*2,3);       % Grid coordinates for each cluster centroid
  clustersFloor{currentFloor}.centroids     = zeros(ncells(1)*ncells(2)*2,nmacs)... % Grid coordinates for each cluster centroid
                                               + newNonDetectedValueProcessed;  
  for i = 1:1:ncells(1)
    
    for j = 1:1:ncells(2)
      
      idxs = ((reducedTrainingLabels(:,1) >= (minCoord(1)+(i-1)*k2)) .*...  % Get fingerprints inside the clusted [ begin ... end [
              (reducedTrainingLabels(:,1) <  (minCoord(1)+(i  )*k2)) .*...
              (reducedTrainingLabels(:,2) >= (minCoord(2)+(j-1)*k2)) .*...
              (reducedTrainingLabels(:,2) <  (minCoord(2)+(j  )*k2)))>0;
              
      if sum(idxs) > 0 % Check if the grid cell contains any fingerprint
        
        % The cell is not empy
        cluster       = cluster +1;
        clusterGlobal = clusterGlobal+1;        
        
        % Extract the rss values and the binary matrix detected/non-detected
        rss   = reducedTrainingMacs(idxs,:);
        valid = reducedTrainingValidMacs(idxs,:);
        
        % The APs that have been heard in the selected fingerprints
        aps   = sum(valid,1)>0;
        
        % Add cluster ID (clusterGlobal) and 
        clustersFloor{currentFloor}.idx(idxs)             = clusterGlobal;
        clustersFloor{currentFloor}.centroids(cluster,aps)= sum(rss(:,aps).*valid(:,aps),1)./sum(valid(:,aps),1);
        clustersFloor{currentFloor}.grid(cluster,:)       = [i,j,currentFloor];        
        
      end
    end
    
  end
  
  for i = 0.5:1:ncells(1)
    
    for j = 0.5:1:ncells(2)
      
      idxs = ((reducedTrainingLabels(:,1) >= (minCoord(1)+(i-1)*k2)) .*...  % Get fingerprints inside the clusted [ begin ... end [
              (reducedTrainingLabels(:,1) <  (minCoord(1)+(i  )*k2)) .*...
              (reducedTrainingLabels(:,2) >= (minCoord(2)+(j-1)*k2)) .*...
              (reducedTrainingLabels(:,2) <  (minCoord(2)+(j  )*k2)))>0;
              
      if sum(idxs) > 0 % Check if the grid cell contains any fingerprint
        
        % The cell is not empy
        cluster       = cluster +1;
        clusterGlobal = clusterGlobal+1;        
        
        % Extract the rss values and the binary matrix detected/non-detected
        rss   = reducedTrainingMacs(idxs,:);
        valid = reducedTrainingValidMacs(idxs,:);
        
        % The APs that have been heard in the selected fingerprints
        aps   = sum(valid,1)>0;
        
        % Add cluster ID (clusterGlobal) and 
        clustersFloor{currentFloor}.idx2(idxs)             = clusterGlobal;
        clustersFloor{currentFloor}.centroids(cluster,aps)= sum(rss(:,aps).*valid(:,aps),1)./sum(valid(:,aps),1);
        clustersFloor{currentFloor}.grid(cluster,:)       = [i,j,currentFloor];        
        
      end
    end
    
  end
  
  
  idxs_newvalues = (clusterGlobal_previous+1):(clusterGlobal);
  if clusterGlobal > clusterGlobal_previous
    clusters.idx(idxsfloor)              = clustersFloor{currentFloor}.idx;
    clusters.idx2(idxsfloor)             = clustersFloor{currentFloor}.idx2;
    clusters.centroids(idxs_newvalues,:) = clustersFloor{currentFloor}.centroids(1:cluster,:);
    clusters.grid(idxs_newvalues,:)      = clustersFloor{currentFloor}.grid(1:cluster,:);
  end

end

clusters.centroids = clusters.centroids(1:clusterGlobal,:);
clusters.grid      = clusters.grid(1:clusterGlobal,:);
clusters.minCoord  = minCoord;
clusters.maxCoord  = maxCoord;
clusters.ncells    = ncells;
##if (nfloors == 1) && (nblds == 1)
##
##  [minCoord] = min(database.trainingLabels(:,1:2));
##  [maxCoord] = max(database.trainingLabels(:,1:2));
##  ncells      = ceil([maxCoord-minCoord]/k2)+1;
##  
##  clusters.idx       = zeros(rsamples,1);
##  clusters.centroids = zeros(ncells(1)*ncells(2),nmacs) + newNonDetectedValueProcessed;
##  clusters.grid      = zeros(ncells(1)*ncells(2),3);
##  
##  cluster = 0;
##  for i = 1:ncells(1)
##    
##    for j = 1:ncells(2)
##      idxs = ((database.trainingLabels(:,1) >= (minCoord(1)+(i-1)*k2)) .*...
##              (database.trainingLabels(:,1) <  (minCoord(1)+(i  )*k2)) .*...
##              (database.trainingLabels(:,2) >= (minCoord(2)+(j-1)*k2)) .*...
##              (database.trainingLabels(:,2) <  (minCoord(2)+(j  )*k2)))>0;
##      if sum(idxs) > 0 
##        cluster = cluster+1;
##        clusters.idx(idxs) = cluster;
##        rss   = database.trainingMacs(idxs,:);
##        valid = database.trainingValidMacs(idxs,:);
##        
##        aps   = sum(valid)>0;
##        clusters.centroids(cluster,aps) = sum(rss(:,aps))./sum(valid(:,aps));
##        clusters.grid(cluster,:)          = [i,j,1];
##        clusters.tempvec(:,cluster) = idxs; % CHECKING
##      end
##    end
##    
##  end
##  clusters.centroids = clusters.centroids(1:cluster,:);
##  clusters.grid      = clusters.grid(1:cluster,:);
##else
##    
##    clusterGlobal = 0;
##    
##    clusters.idx       = zeros(rsamples,1);
##    clusters.centroids = [];
##    clusters.grid      = [];
##    for currentFloor = 1:nfloors
##    
##      idxsfloor                = (database.trainingLabels(:,4) == currentFloor);
##      reducedTrainingLabels    = database.trainingLabels(idxsfloor,:);
##      reducedTrainingMacs      = database.trainingMacs(idxsfloor,:);
##      reducedTrainingValidMacs = database.trainingValidMacs(idxsfloor,:);
##      
##      rsamplesReduced = size(reducedTrainingLabels,1);
##      [minCoord] = min(reducedTrainingLabels(:,1:2));
##      [maxCoord] = max(reducedTrainingLabels(:,1:2));
##      ncells     = ceil([maxCoord-minCoord]/k2)+1;
##      
##      clustersFloor{currentFloor}.idx           = zeros(rsamplesReduced,1);
##      clustersFloor{currentFloor}.grid          = zeros(ncells(1)*ncells(2),3);
##      clustersFloor{currentFloor}.centroids     = zeros(ncells(1)*ncells(2),nmacs) + newNonDetectedValueProcessed;
##      
##      cluster = 0;
##      for i = 1:ncells(1)
##        
##        for j = 1:ncells(2)
##          idxs = ((reducedTrainingLabels(:,1) >= (minCoord(1)+(i-1)*k2)) .*...
##                  (reducedTrainingLabels(:,1) <  (minCoord(1)+(i  )*k2)) .*...
##                  (reducedTrainingLabels(:,2) >= (minCoord(2)+(j-1)*k2)) .*...
##                  (reducedTrainingLabels(:,2) <  (minCoord(2)+(j  )*k2)))>0;
##          if sum(idxs) > 0 
##            
##            cluster = cluster +1;
##            clusterGlobal = clusterGlobal+1;
##            
##            
##            rss   = reducedTrainingMacs(idxs,:);
##            valid = reducedTrainingValidMacs(idxs,:);
##            
##            aps   = sum(valid,1)>0;
##            clustersFloor{currentFloor}.idx(idxs) = clusterGlobal;
##            clustersFloor{currentFloor}.centroids(cluster,aps)= sum(rss(:,aps))./sum(valid(:,aps));
##            clustersFloor{currentFloor}.grid(cluster,:)       = [i,j,currentFloor];
##            clustersFloor{currentFloor}.tempvec(:,cluster) = idxs; % CHECKING
##          end
##        end
##        
##      end
##      clusters.idx(idxsfloor) = clustersFloor{currentFloor}.idx;
##      clusters.centroids = [clusters.centroids;clustersFloor{currentFloor}.centroids(1:cluster,:)];
##      clusters.grid      = [clusters.grid;clustersFloor{currentFloor}.grid(1:cluster,:)];
##      %clusters.centroids = 
##    end
##end
%clusters = smallestEnclosingClircle(database.trainingLabels, 1);
##[clusters.idx, clusters.centroids, clusters.sumd, clusters.dist] = kmeans_local (database.trainingMacs, k2 ,'distance', 'cityblock','emptyaction', 'singleton','start','plus');
##csamples = size(clusters.centroids,1);


csamples = max(max(clusters.idx),max(clusters.idx2));
vecidxcsamples = 1:csamples;
idxs     = cell(1,csamples);
for i = 1:csamples
  idxs(i) = unique([vecidxsamples(clusters.idx == i),vecidxsamples(clusters.idx2 == i)]); 

  %i_1 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[ 0,+1,0]),2)==3);
  %i_2 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[ 0,-1,0]),2)==3); 
  %i_3 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[+1,0,0]),2)==3);
  %i_4 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[-1,0,0]),2)==3);
  %i_5 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[+1,+1,0]),2)==3);
  %i_6 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[-1,+1,0]),2)==3);
  %i_7 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[+1,-1,0]),2)==3);
  %i_8 = vecidxcsamples(sum(clusters.grid == (clusters.grid(i,:)+[-1,-1,0]),2)==3); 

  %idxs(i) = vecidxsamples(sum(clusters.idx == [i,i_1,i_2,i_3,i_4,i_5,i_6,i_7,i_8],2)>0); 
  
end

%if (sum(sum(clusters.tempvec,1)) ~= rsamples) || (sum(sum(clusters.tempvec,2) ~= 1)>0)
%  disp('Error: number of samples do not match');
%  pause;  
%end
time00 = myToc(ticT1m3);

fprintf('\nGrid-based Clustering finished\n')
fprintf('    radious for size       : %d\n',k2)
fprintf('    number of clusters     : %d\n\n',csamples)

for i = 1:osamples    

    time01 = myToc(ticT1m3);
        
    ofp = database.testMacs(i,:);
	
    dist_cluster = zeros(1,csamples);
		%% Find Cluster
    for j = 1:csamples
      dist_cluster(1,j) = distance_cityblock(clusters.centroids(j,:),ofp);
    end
    [value,cluster] = min(dist_cluster);
    
	  samples = unique(cell2mat(idxs(cluster)));
	
    if size(samples,1)==0
      samples = 1:rsamples;      
    end
  
	  trainingMacs      = database.trainingMacs(samples,:);
	  trainingValidMacs = database.trainingValidMacs(samples,:);	
	  trainingLabels    = database.trainingLabels(samples,:);
	
	  rsamplesreduced = size(trainingMacs,1);
    
    ofp = database.testMacs(i,:);
    
    time02 = myToc(ticT1m3);
    
    distances = zeros(1,rsamplesreduced);
 
    for j = 1:rsamplesreduced;
        distances(j) = feval(distanceMetric,trainingMacs(j,:),ofp,additionalparams);
    end
           
    time03 = myToc(ticT1m3);
    % Sort distances
    [distancessort,candidates] = sort(distances);
    
    % Set the value of candidates to k (kNN)  <---- THIS IS NEW!!!!!!
    % MENTION IN THE PAPER
    ncandidates = min(k,rsamplesreduced);
    
    % Check if candidates ranked in positions higher than k have the same
    % distance in the feature space.
    while ncandidates < rsamplesreduced 
       if abs(distancessort(ncandidates)-distancessort(ncandidates+1))<0.000000000001
           ncandidates = ncandidates+1;
       else
           break
       end
    end

    % Estimate the building from the ncandidates
    probBld = zeros(1,nblds);
    for bld = 1:nblds
        probBld(bld) = sum(trainingLabels(candidates(1:ncandidates),5)==bld);
    end
    [~,bld] = max(probBld);
    
    % Estimate the floor from the ncandidates
    probFloor = zeros(1,nfloors);
    for floor = 1:nfloors
        probFloor(floor) = sum( (trainingLabels(candidates(1:ncandidates),4)==floor).*...
                                (trainingLabels(candidates(1:ncandidates),5)==bld));
    end
    [~,floor] = max(probFloor);
    
    % Estimate the coordinates from the ncandidates that belong to the
    % estimated floor
    points = 0;
    point  = [0,0,0];
    for j = 1:ncandidates
        if (trainingLabels(candidates(j),4)==floor).*(trainingLabels(candidates(j),5)==bld)
            points = points + 1;
            point  = point + trainingLabels(candidates(j),1:3);
        end
    end
    point = point / points;
    
    time04 = myToc(ticT1m3);
    % Extract the real-world bld (revert bld remap)
    realWorldBld = remapVector(bld,1:nblds,origBlds');
    
    % Extract the real-world floor (revert floor remap)
    realWorldFloor = remapVector(floor,1:nfloors,origFloors');
    
    % Generate the results statistics
    results.error(i,1) = distance_euclidean(database.testLabels(i,1:2),point(1:2)); % 2D Positioning error in m 
    results.error(i,2) = distance_euclidean(database.testLabels(i,1:3),point(1:3)); % 3D Positioning error in m
    results.error(i,3) = abs(database.testLabels(i,3)-point(3));                    % Height difference in m
    results.error(i,4) = abs(realWorldFloor - database_orig.testLabels(i,4));       % Height difference in floors
    results.error(i,5) = (realWorldBld ~= database_orig.testLabels(i,5));           % Building error
    
    results.prediction(i,:) = [point,realWorldFloor,realWorldBld];                  % Predicted position [x,y,z,floor]
    results.targets(i,:)    = database_orig.testLabels(i,1:5);                      % Current position   [x,y,z,floor]
    results.candidates(i,1) = ncandidates;                                          % Number of nearest neighbours used to estimate the position
    results.distances(i,1)  = distancessort(1);                                     % Distance in feature space of the best match
    
    results.samples(i,1)    = rsamplesreduced;         % Distance in feature space of the best match
    results.timesample(i,1) = time01;                  % Distance in feature space of the best match
    results.timesample(i,2) = time02;                  % Distance in feature space of the best match
    results.timesample(i,3) = time03;                  % Distance in feature space of the best match    
    results.timesample(i,4) = time04;                  % Distance in feature space of the best match    
    results.timesample(i,5) = myToc(ticT1m3);                     % Distance in feature space of the best match    
    
    time05 = myToc(ticT1m3);
	
end

results.timefull = [time00pre,time00,myToc(ticT1m3)];

fprintf('Successfully executed!\n')
fprintf('    Mean Positioning Error 2D*  : %3.3f\n',mean(results.error((results.error(:,4)==0),1)));  % Mean 2D Positioning error when estimate and true position are in the same floor
fprintf('    Mean Positioning Error 3D   : %3.3f\n',mean(results.error(:,2)));   
fprintf('    Mean Positioning Error EvAAL: %3.3f\n',mean(results.error(:,1)+4*results.error(:,4)+50*results.error(:,5)));                        % Mean 3D Positioning error
fprintf('    Mean Positioning Error IPIN : %3.3f\n',mean(results.error(:,1)+15*results.error(:,4)+50*results.error(:,5)));                        % Mean 3D Positioning error
fprintf('    Bld&Floor detection hit rate: %3.3f\n',mean((results.error(:,4)==0).*(results.error(:,5)==0))*100);                 % Floor detection rate
fprintf('    BD Preprocessing time       : %3.6f\n',time00);    
fprintf('    Average preprocessing time  : %3.6f\n',mean(results.timesample(:,2)-results.timesample(:,1)));    
fprintf('    Average matching time       : %3.6f\n',mean(results.timesample(:,3)-results.timesample(:,2)));    
fprintf('    Average positioning time    : %3.6f\n',mean(results.timesample(:,4)-results.timesample(:,3)));   
fprintf('    Full time                   : %3.6f\n',results.timefull(3)); 
fprintf('        DB Preprocessing        : %3.3f%%\n',results.timefull(1)/results.timefull(3)*100); 
fprintf('        FP Preprocessing        : %3.3f%%\n',sum(results.timesample(:,2)-results.timesample(:,1))/results.timefull(3)*100); 
fprintf('        Matching                : %3.3f%%\n',sum(results.timesample(:,3)-results.timesample(:,2))/results.timefull(3)*100); 
fprintf('        Positioning             : %3.3f%%\n\n',sum(results.timesample(:,4)-results.timesample(:,3))/results.timefull(3)*100); 